package luchr.io;

import luchr.lang.Collection;

import java.io.File;
import java.io.BufferedReader;
import java.io.FileReader;
import java.io.FileWriter;

import java.util.Enumeration;

public class IniFile
{
	static protected class RTException extends java.lang.RuntimeException
	{
		public RTException () {super();}
		public RTException (String message) {super(message);}
	}

	static public class Entry implements java.io.Serializable
	{
		static public interface Controller
		{
			public boolean acceptClient (Entry e, boolean inInit);
			public void setupClient (Entry e, boolean inInit);
			public void releaseClient (Entry e);
			public void clientOwnerHasChanged (Section oldOwner, Section newOwner);
			public boolean canClientChangeValue (Entry e, String newValue);
			public boolean canClientChangeKey (Entry e, String newKey);
			public boolean canClientChangeController
				(Controller oldController, Controller newController);
		}

		static public class Adapter implements Controller
		{
			public boolean acceptClient (Entry e, boolean inInit)
				{return true;}
			public void setupClient (Entry e, boolean inInit) {;}
			public void releaseClient (Entry e) {;}
			public void clientOwnerHasChanged (Section oldOnwer, Section newOwner)
				{;}
			public boolean canClientChangeValue (Entry e, String newValue)
				{return true;}
			public boolean canClientChangeKey (Entry e, String newKey)
				{return true;}
			public boolean canClientChangeController
				(Controller oldController, Controller newController)
				{return true;}
		}

		static public class ChangeException extends RTException
		{
			public ChangeException () {super();}
			public ChangeException (String message) {super(message);}
		}


		private String			key, value;
		private Controller	controller			= null;
		private Section		 owner					 = null;

		public Entry (String key, String value, Controller controller)
		{
			super();
			setController(controller,true);
			setKey(key);
			setValue(value);
		}
		public Entry (String key, String value) {this(key,value,null);}
		public Entry (String key) {this(key,"");}
		public Entry (Controller controller, String key, String value)
			{this(key,value,controller);}

		public java.lang.String toString ()
		{
			return "IniFile.Entry: "+key+"="+value;
		}

		/* getKey */
		public final String getKey () {return key;}

		/* getxxxValue */
		public final String getStringValue ()
		{
			if (value==null) return new String(""); else
			return value;
		}
		public final byte getByteValue () throws NumberFormatException
			{return Byte.parseByte(value);}
		public final short getShortValue () throws NumberFormatException
			 {return Short.parseShort(value);}
		public final int getIntValue () throws NumberFormatException
			{return Integer.parseInt(value);}
		public final long getLongValue () throws NumberFormatException
			{return Long.parseLong(value);}
		public final float getFloatValue () throws NumberFormatException
			{return Float.valueOf(value).floatValue();}
		public final double getDoubleValue () throws NumberFormatException
			{return Double.valueOf(value).doubleValue();}
		public final boolean getBooleanValue () throws IllegalArgumentException
		{
			if (value!=null)
			{
				if (value.equalsIgnoreCase("TRUE")) return true;
				if (value.equalsIgnoreCase("FALSE")) return false;
			}
			throw new IllegalArgumentException("That was not a boolean!");
		}

		/* getOwner */
		public final Section getOwner () {return owner;}

		/* getController */
		public final Controller getController () {return controller;}

		/* setOwner */
		final void setOwner (Section owner)
		{
			Section oo = this.owner;
			this.owner = owner;
			if (controller!=null) controller.clientOwnerHasChanged(oo,owner);
		}

		/* setController */
		private final void setController (Controller controller,
																			boolean inInit) throws
			ChangeException
		{
			if (this.controller!=null &&
				!this.controller.canClientChangeController(
					 this.controller, controller))
				throw new ChangeException("Controller does not allow to "+
					"change controller!");
			if (controller!=null && !controller.acceptClient(this,inInit))
				throw new ChangeException("New controller ("+
					controller.toString()+") rejects client ("+
					this.toString()+")!");
			if (this.controller!=null) this.controller.releaseClient(this);
			this.controller = controller;
			if (controller!=null) controller.setupClient(this,inInit);
		}
		public final void setController (Controller controller) throws
			ChangeException
			{setController(controller,false);}

		/* setKey */
		public void setKey (String key)
			throws ChangeException
		{
			if (key==null || key.equals(""))
				throw new ChangeException("Invalid entry key");
			if (owner!=null && owner.hasEntry(key))
				throw new ChangeException("Duplicate entry key; "+
					"cannot change entry\n"+toString()+"\nto key: "+key);
			if (controller!=null && !controller.canClientChangeKey(this,key))
				throw new ChangeException("Controller does not allow to "+
					"change entry!\n"+toString()+"\nnew key: "+key);
			this.key = key;
		}

		/* setValue */
		public void setValue (String value) throws
			ChangeException
		{
			if (value==null) throw new ChangeException("Invalid value");
			if (controller!=null && !controller.canClientChangeValue(this,value))
				throw new ChangeException("Controller does not allow to "+
					"change entry!\n"+toString()+"\nnew value: "+value); 
			this.value = value;
		}
		public final void setValue (int i) {setValue(String.valueOf(i));}
		public final void setValue (long l) {setValue(String.valueOf(l));}
		public final void setValue (float f) {setValue(String.valueOf(f));}
		public final void setValue (double d) {setValue(String.valueOf(d));}
		public final void setValue (char c) {setValue(String.valueOf(c));}
		public final void setValue (boolean b) {setValue(String.valueOf(b));}
		public final void setValue (Object o) {setValue(String.valueOf(o));}

		/* debug */
		public void debug ()
		{
			System.out.println(toString());
			System.out.println("Controller: "+
				(controller==null?"null":controller.toString()));
		}

		/* writeEntry */
		public void writeEntry (final BufferedWriter bw) throws
			java.io.IOException
			{bw.writeln(getKey()+"="+getStringValue());}
	}

	static public class Section implements java.io.Serializable
	{
		static public interface Controller
		{
			public boolean acceptClient (Section s, boolean inInit);
			public void setupClient (Section s, boolean inInit);
			public void releaseClient (Section s);
			public void clientOwnerHasChanged (Manager oldManager,
				Manager newManager);
			public boolean canClientInsertEntry (Section s, Entry e);
			public boolean canClientDeleteEntry (Section s, Entry e);
			public boolean canClientChangeName (Section s, String newName);
			public boolean canClientChangeController
				(Controller oldController, Controller newController);
		}

		static public class Adapter implements Controller
		{
			public boolean acceptClient (Section s, boolean inInit)
				{return true;}
			public void setupClient (Section s, boolean inInit) {;}
			public void releaseClient (Section s) {;}
			public void clientOwnerHasChanged (Manager oldManager,
				Manager newManager) {;}
			public boolean canClientInsertEntry (Section s, Entry e)
				{return true;}
			public boolean canClientDeleteEntry (Section s, Entry e)
				{return true;}
			public boolean canClientChangeName (Section s, String newName)
				{return true;}
			public boolean canClientChangeController
				(Controller oldController, Controller newController)
				{return true;}
		}

		static public class InsertEntryException extends RTException
		{
			public InsertEntryException () {super();}
			public InsertEntryException (String message) {super(message);}
		}

		static public class DeleteEntryException extends RTException
		{
			public DeleteEntryException () {super();}
			public DeleteEntryException (String message) {super(message);}
		}

		static public class ChangeException extends RTException
		{
			public ChangeException () {super();}
			public ChangeException (String message) {super(message);}
		}

		private	 String					 name;
		protected Collection			 entryColl;
		private	 Controller			 controller;
		private	 Manager					owner				 =null;

		public Section (Controller controller, String name,
										int capacity, int delta)
		{
			super();
			setController(controller,true);
			setName(name);
			entryColl = new Collection (capacity,delta);
		}
		public Section (String name, int capacity, int delta)
			{this(null,name,capacity,delta);}
		public Section (String name, int capacity) {this(name,capacity,0);}
		public Section (String name) {this(name,5);}
		public Section () {this("");}
		public Section (String name, int capacity, int delta,
										Controller controller)
			{this(controller,name,capacity,delta);}
		
		public java.lang.String toString ()
		{
			return "IniFile.Section: "+name;
		}

		/* getOwner */
		public final Manager getOwner () {return owner;}

		/* getName */
		public final String getName () {return name;}

		/* getController */
		public final Controller getController () {return controller;}

		/* getEntryByKey */
		public final Entry getEntryByKey (final String key)
		{
			if (key==null) return null;
			return ((Entry)entryColl.firstThat(new Collection.Iteration () {
				public boolean doComparison (Object item)
					{return key.equalsIgnoreCase(((Entry)item).getKey());}
			}));
		}

		/* getFirstEntryByValue */
		public final Entry getFirstEntryByValue (final String value)
		{
			if (value==null) return null;
			return ((Entry)entryColl.firstThat(new Collection.Iteration () {
				public boolean doComparison (Object item)
					{return value.equals(((Entry)item).getStringValue());}
			}));
		}

		/* getLastEntryByValue */
		public final Entry getLastEntryByValue (final String value)
		{
			if (value==null) return null;
			return ((Entry)entryColl.lastThat(new Collection.Iteration () {
				public boolean doComparison (Object item)
					{return value.equals(((Entry)item).getStringValue());}
			}));
		}

		/* getEntryEnumerator */
		public Enumeration getEntryEnumerator () {return entryColl.items();}

		/* getEntryCollectionClone */
		public Collection getEntryCollectionClone ()
			{return (Collection)(entryColl.clone());}

		/* debug */
		public void debug ()
		{
			System.out.println(toString());
			System.out.println("Controller: "+
				(controller==null?"null":controller.toString()));
			entryColl.debug();
		}

		/* setOwner */
		final void setOwner (Manager owner)
		{
			Manager oo = this.owner;
			this.owner = owner;
			if (controller!=null) controller.clientOwnerHasChanged(oo,owner);
		}

		/* setName */
		public void setName (String name) throws
			ChangeException
		{
			if (name==null || name.equals(""))
				throw new ChangeException("Invalid section name");
			if (owner!=null && owner.hasSection(name))
				throw new ChangeException("Duplicate section name; "+
					"cannot change section\n"+toString()+"\nto name: "+name);
			if (controller!=null && !controller.canClientChangeName(this,name))
				throw new ChangeException("Controller does not allow to change name!\n"+
					toString()+"\nnew name: "+name);
			this.name = name;
		}

		/* setController */
		private final void setController (Controller controller,
																			boolean inInit) throws
			ChangeException
		{
			if (this.controller!=null &&
					!this.controller.canClientChangeController(
						this.controller, controller))
				throw new ChangeException("Controller does not allow to "+
					"change controller!");
			if (controller!=null && !controller.acceptClient(this,inInit))
				throw new ChangeException("New controller ("+
					controller.toString()+") rejects client ("+
					this.toString()+")!");
			if (this.controller!=null) this.controller.releaseClient(this);
			this.controller = controller;
			if (controller!=null) controller.setupClient(this,inInit);
		}
		public final void setController (Controller controller) throws
			ChangeException
			{setController(controller,false);}

		/* hasEntry */
		public final boolean hasEntry (String key)
			{return getEntryByKey(key)!=null;}

		/* insertEntry */
		public void insertEntry (Entry e) throws
			InsertEntryException
		{
			if (e==null)
				throw new InsertEntryException("Cannot insert \"null\"-entry!");
			if (hasEntry(e.getKey()))
				throw new InsertEntryException("Duplicate entry key; "+
					"cannot insert entry\n"+e.toString());
			if (e.getOwner()!=null)
				throw new InsertEntryException("Cannot insert entry: "+e.toString()+
					", because entry belongs already to an other section!");
			if (controller!=null && !controller.canClientInsertEntry(this,e))
				throw new InsertEntryException("Controller does not allow to "+
					"insert entry!\n"+e.toString());
			try
			{
				entryColl.insert(e);
				e.setOwner(this);
			} catch (Collection.ObjectAlreadyInCollException ex)
				{ // this cannot happen
					ex.printStackTrace();
				}
		}

		/* deleteEntry */
		public void deleteEntry (Entry e) throws
			DeleteEntryException
		{
			if (e==null)
				throw new DeleteEntryException("Cannot find entry: \"null\"");
			if (!hasEntry(e.getKey()))
				throw new DeleteEntryException("Cannot find entry: "+e.getKey());
			if (controller!=null && !controller.canClientDeleteEntry(this,e))
				throw new DeleteEntryException("Controller does not allow to "+
					"delete entry!\n"+e.toString());
			try
			{
				entryColl.delete(e);
				e.setOwner(null);
			} catch (Collection.ObjectNotFoundException ex)
				{ // this cannot happen
					ex.printStackTrace();
				}
		}

		/* deleteAllEntries */
		public void deleteAllEntries ()
			throws DeleteEntryException
		{
			entryColl.forEach(new Collection.Iteration () {
				public void doAction (Object item)
					{deleteEntry(((Entry)item));}
			});
		}

		/* writeSection */
		public void writeSection (final BufferedWriter bw) throws
			java.io.IOException
		{
			bw.writeln("["+getName()+"]");
			Enumeration enum=entryColl.items();
			while (enum.hasMoreElements())
				((Entry)(enum.nextElement())).writeEntry(bw);
		}
	}

	static public class Manager implements java.io.Serializable
	{
		static public class ChangeException extends RTException
		{
			public ChangeException () {super();}
			public ChangeException (String message) {super(message);}
		}

		static public interface Controller
		{
			public boolean acceptClient (Manager m, boolean inInit);
			public void setupClient (Manager m, boolean inInit);
			public void releaseClient (Manager m);
			public boolean canClientInsertSection (Manager m, Section s);
			public boolean canClientDeleteSection (Manager m, Section s);
			public boolean canClientChangeFile (Manager m, File newFile);
			public boolean canClientChangeController
				(Controller oldController, Controller newController);
		}

		static public class Adapter implements Controller
		{
			public boolean acceptClient (Manager m, boolean inInit)
				{return true;}
			public void setupClient (Manager m, boolean inInit) {;}
			public void releaseClient (Manager m) {;}
			public boolean canClientInsertSection (Manager m, Section s)
				{return true;}
			public boolean canClientDeleteSection (Manager m, Section s)
				{return true;}
			public boolean canClientChangeFile (Manager m, File newFile)
				{return true;}
			public boolean canClientChangeController
				(Controller oldController, Controller newController)
				{return true;}
		}

		static public class InsertSectionException extends RTException
		{
			public InsertSectionException () {super();}
			public InsertSectionException (String message) {super(message);}
		}

		static public class DeleteSectionException extends RTException
		{
			public DeleteSectionException () {super();}
			public DeleteSectionException (String message) {super(message);}
		}

		private class Line
		{
			String				 text,orig;

			Line (BufferedReader reader)
				throws java.io.IOException
			{
				super();
				orig = reader.readLine();
				text = orig.trim();
			}

			boolean isSection ()
			{
				return text.length()>2 && text.startsWith("[") && text.endsWith("]");
			}

			String getSectionName ()
			{
				if (!isSection()) return null;
				return text.substring(1,text.length()-1);
			}

			boolean isValidEntry () {return text.indexOf("=")>0;}

			String getFront ()
			{
				if (!isValidEntry()) return null;
				return text.substring(0,text.indexOf("="));
			}

			String getBack ()
			{
				if (!isValidEntry()) return null;
				return text.substring(text.indexOf("=")+1);
			}

			boolean isEmpty() {return text.length()==0;}

			boolean isComment ()
			{
				if (isEmpty()) return false;
				String upText=text.toUpperCase();
				return (upText.startsWith(";") || upText.startsWith("REM"));
			}
		}

		protected Collection secColl;
		protected File			 iniFile;
		private	 Controller controller;
		public		boolean		overrideDuplicateEntries	= false;
		public		boolean		overrideDuplicateSections = false;

		public Manager (File iniFile, int capacity, int delta,
										Controller controller)
		{
			super();
			setController(controller,true);
			setFile(iniFile);
			secColl = new Collection(capacity,delta);
		}
		public Manager (File iniFile, int capacity, int delta)
			{this(iniFile,capacity,delta,null);}
		public Manager (File iniFile, int capacity)
			{this(iniFile,capacity,5);}
		public Manager (File iniFile) {this(iniFile,5);}

		public java.lang.String toString ()
		{
			return "IniFile.Manager: "+iniFile.toString();
		}

		/* getFile */
		public final File getFile () {return iniFile;}

		/* getController */
		public final Controller getController () {return controller;}

		/* getSection */
		public final Section getSection (final String name)
		{
			if (name==null) return null;
			return ((Section)secColl.firstThat(new Collection.Iteration () {
				public boolean doComparison (Object item)
					{return name.equalsIgnoreCase(((Section)item).getName());}
			}));
		}

		/* getSectionEnumerator */
		public Enumeration getSectionEnumerator () {return secColl.items();}

		/* getSectionCollectionClone */
		public Collection getSectionCollectionClone ()
			{return (Collection)(secColl.clone());}
								 
		/* getNewSection */
		public Section getNewSection (String name) {return new Section(name);}

		/* getNewEntry */
		public Entry getNewEntry (Section s, String key, String value)
			{return new Entry(key,value);}

		/* debug */
		public void debug ()
		{
			System.out.println(toString());
			System.out.println("Controller: "+
				(controller==null?"null":controller.toString()));
			secColl.debug();
		}

		/* setFile */
		public void setFile (File iniFile) throws
			ChangeException
		{
			if (iniFile==null) throw new ChangeException("Invalid file");
			if (controller!=null &&
					 !controller.canClientChangeFile(this,iniFile))
				throw new ChangeException("Controller does not allow to change "+
					"file!");
			this.iniFile = iniFile;
		}

		/* setController */
		private final void setController (Controller controller,
																			boolean inInit) throws
			ChangeException
		{
			if (this.controller!=null &&
					!this.controller.canClientChangeController(
						this.controller, controller))
				throw new ChangeException("Controller does not allow to "+
					"change controller!");
			if (controller!=null && !controller.acceptClient(this,inInit))
				throw new ChangeException("New controller ("+
					controller.toString()+") rejects client ("+
					this.toString()+")!");
			if (this.controller!=null) this.controller.releaseClient(this);
			this.controller = controller;
			if (controller!=null) controller.setupClient(this,inInit);
		}
		public final void setController (Controller controller) throws
			ChangeException
			{setController(controller,false);}

		/* hasSection */
		public final boolean hasSection (String name)
			{return getSection(name)!=null;}

		/* insertSection */
		public void insertSection (Section s) throws
			InsertSectionException
		{
			if (s==null)
				throw new InsertSectionException("Cannot insert \"null\"-section!");
			if (hasSection(s.getName()))
				throw new InsertSectionException("Duplicate section; "+
					"cannot insert section\n"+s.toString());
			if (s.getOwner()!=null)
				throw new InsertSectionException("Cannot insert section: "+
					s.toString()+"because section belongs already to an other Manager");
			if (controller!=null && !controller.canClientInsertSection(this,s))
				throw new InsertSectionException("Controller does not allow to "+
					"insert section!\n"+s.toString());
			try
			{
				secColl.insert(s);
				s.setOwner(this);
			} catch (Collection.ObjectAlreadyInCollException ex)
				{ // this cannot happen
					ex.printStackTrace();
				}
		}

		/* deleteSection */
		public void deleteSection (Section s) throws
			DeleteSectionException
		{
			if (s==null)
				throw new DeleteSectionException("Cannot find section: \"null\"");
			if (!hasSection(s.getName()))
				throw new DeleteSectionException("Cannot find section: "+s.getName());
			if (controller!=null && !controller.canClientDeleteSection(this,s))
				throw new DeleteSectionException("Controller does not allow to "+
					"delete section!\n"+s.toString());
			try
			{
				secColl.delete(s);
				s.setOwner(null);
			} catch (Collection.ObjectNotFoundException ex)
				{ // this cannot happen
					ex.printStackTrace();
				}
		}

		/* deleteAllSections */
		public void deleteAllSections ()
			throws DeleteSectionException
		{
			secColl.forEach(new Collection.Iteration () {
				public void doAction (Object item) {deleteSection(((Section)item));}
			});
		}

		/* lineWithoutSection */
		public void lineWithoutSection (String line)
			{System.err.println("lineWithoutSection: "+line);}

		/* notValidEntry */
		public void notValidEntry (String line)
			{System.err.println("notValidEntry: "+line);}

		/* stopAfterInsertSectionException */
		public boolean stopAfterInsertSectionException (Section s) {return true;}

		/* stopAfterInsertEntryException */
		public boolean stopAfterInsertEntryException (Section s, Entry e)
			{return true;}

		/* readFile */
		public void readFile () throws
			java.io.FileNotFoundException,
			java.io.IOException
		{
			final BufferedReader br = new BufferedReader(new
				FileReader(iniFile));
			deleteAllSections();
			Section currSec = null;
			Line l;
			boolean doInsert;
			try
			{
				while (br.ready())
				{
					l = new Line(br);
					if (l.isEmpty() || l.isComment()) continue;
					if (l.isSection())
					{
						String secName=l.getSectionName();
						doInsert=true;
						if (hasSection(secName))
						{
							if (overrideDuplicateSections)
								deleteSection(getSection(secName));
								else doInsert=false;
						}

						if (doInsert)
						{
							currSec = getNewSection(secName);
							try {insertSection(currSec);}
							catch (InsertSectionException ex)
								{if (stopAfterInsertSectionException(currSec)) throw ex;}
						} else currSec = new Section("Temp");
							
					} else
						if (currSec==null) lineWithoutSection(l.orig); else
						{
							if (l.isValidEntry())
							{
								final String entKey=l.getFront();
								final String entVal=l.getBack();
								doInsert=true;
								if (currSec.hasEntry(entKey))
								{
									if (overrideDuplicateEntries)
										currSec.deleteEntry(currSec.getEntryByKey(entKey));
										else doInsert=false;
								}
								
								if (doInsert)
								{
									Entry currEnt=getNewEntry(currSec,entKey,entVal);	
									try {currSec.insertEntry(currEnt);}
									catch (Section.InsertEntryException ex)
										{
											if (stopAfterInsertEntryException(currSec,currEnt))
												throw ex;
										}
								}
							} else notValidEntry(l.orig);
						}
				}
			} finally {br.close();}
		}

		/* writeFile */
		public void writeFile () throws
			java.io.IOException
		{
			boolean IOException=false;
			final BufferedWriter bw = new BufferedWriter(new
				FileWriter(iniFile));
			try
			{
				Enumeration enum=secColl.items();
				while (enum.hasMoreElements())
				{
					((Section)(enum.nextElement())).writeSection(bw);
					bw.newLine();
				}
			} finally {bw.flush();bw.close();}
		}
	}

	static public class EntryChecker
	{
		private EntryChecker () {super();}

		protected static Entry getNewEntry (String key, String value)
		{
			return new Entry(key,value);
		}
		public static void checkIntEntry(Section sec, String entryKey,
															int defaultValue)
		{
			if (!sec.hasEntry(entryKey))
			{
				sec.insertEntry(getNewEntry(entryKey,String.valueOf(defaultValue)));
				return;
			}
			Entry e = sec.getEntryByKey(entryKey);
			try
			{
				e.getIntValue();
			} catch (NumberFormatException ex)
				{
					e.setValue(defaultValue);
				}
		}

		public static void checkStringEntry(Section sec, String entryKey,
																 String defaultValue)
		{
			if (!sec.hasEntry(entryKey))
			{
				sec.insertEntry(getNewEntry(entryKey,defaultValue));
				return;
			}
			if (defaultValue.equals("")) return;
			Entry e = sec.getEntryByKey(entryKey);
			if (e.getStringValue().equals("")) e.setValue(defaultValue);
		}
	}
}
